#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
=======
SeisLib
=======
"""

from __future__ import print_function
import sys
import os
from .__version__ import __version__
from . import an
from . import colormaps
from . import eq
from . import exceptions
from . import plotting
from . import tomography
from . import utils



